# -*- coding: utf-8 -*-
import logging
import sys


class ParnasLogger(logging.Logger):
    PLAIN = 55  # Add a new logging level.

    def plain(self, message, *args, **kws):
        self.log(self.PLAIN, message, *args, **kws)
logging.addLevelName(ParnasLogger.PLAIN, 'PLAIN')


class LogFormatter(logging.Formatter):
    """
    Adapted from https://stackoverflow.com/a/62488520.
    """

    def __init__(self):
        super().__init__(datefmt="%H:%M:%S")

    def format(self, record):
        if record.levelno == ParnasLogger.PLAIN:
            self._style._fmt = "%(message)s"
        else:
            if record.levelno == logging.INFO:
                self._style._fmt = "[parnas %(asctime)s] %(message)s"
            else:
                color = {
                    logging.WARNING: 33,
                    logging.ERROR: 31,
                    logging.FATAL: 31,
                    logging.DEBUG: 36
                }.get(record.levelno, 0)
                self._style._fmt = f"[\033[{color}m%(levelname)s\033[0m] %(message)s"
        return super().format(record)


logging.setLoggerClass(ParnasLogger)
parnas_logger = logging.getLogger('Parnas logger')
formatter = LogFormatter()
handler = logging.StreamHandler(sys.stderr)
handler.setFormatter(formatter)

parnas_logger.addHandler(handler)
parnas_logger.setLevel(logging.INFO)



# -*- coding: utf-8 -*-
from collections import deque
from typing import Dict, List

import numpy as np
from dendropy import Tree, Node
from datetime import datetime

from .medoid_utils import DistFunction
from .pmedian_utils import cost_function, filtered_preorder_iterator, filtered_postorder_iterator, dfs_tree_traversal
from .tree_indexer import TreeIndexer
from parnas.logging import parnas_logger


class PMedianFinder(object):
    """
    Class that finds the median nodes (among leaves) given a phylogenetic tree.
    An adapted implementation of Tamir's algorithm (Tamir 1996) for generalized p-medians.
    """

    def __init__(self, tree: Tree):
        tree_indexer = TreeIndexer(tree.taxon_namespace)
        tree_indexer.index_tree(tree)  # Now all nodes in the tree have an 'index' field. First n indices = leaves.
        self.cost_function = cost_function

        self.nodes = list(tree.nodes())
        self.nnodes = len(self.nodes)
        self.nleaves = len(tree.leaf_nodes())
        self.tree = tree

        self.distance_lookup = np.zeros((self.nnodes, self.nleaves), dtype=np.float64)
        self.index_lookup = np.zeros((self.nnodes, self.nleaves), dtype=np.int32)
        self.is_ancestor = np.full((self.nnodes, self.nnodes), False, dtype=np.bool_)
        self.subtree_leaves = np.zeros(self.nnodes, dtype=np.int32)
        self.children = np.full((self.nnodes, 2), -1, dtype=np.int32)
        self.leaf_lists = np.zeros((self.nnodes, self.nleaves), dtype=np.int32)

        self.n_c = len(self.tree.leaf_nodes())
        self.G = np.array([], dtype=np.float64)
        self.F = np.array([], dtype=np.float64)

    def find_medoids(self, p: int, distance_functions: Dict[Node, DistFunction]):
        parnas_logger.debug(f'Started find_medoids {datetime.now().strftime("%H:%M:%S")}')
        self.distance_functions = distance_functions
        self.n_c = min(p, len(self.tree.leaf_nodes()))
        self.G = np.full((self.n_c + 1, self.nnodes, self.nleaves), np.inf, dtype=np.float64)
        self.F = np.full((self.n_c + 1, self.nnodes, self.nleaves), np.inf, dtype=np.float64)
        parnas_logger.debug(f'Allocated arrays {datetime.now().strftime("%H:%M:%S")}')

        self._initialize_lookups()
        parnas_logger.debug(f'Finished initialization {datetime.now().strftime("%H:%M:%S")}')

        processed_nodes = 0
        i = 0
        for node in self.tree.postorder_node_iter():
            if node.is_leaf():
                self._initialize_G_and_F(node)
            else:
                for q in range(min(self.subtree_leaves[node.index], self.n_c) + 1):
                    for radius_id in self.leaf_lists[node.index]:
                        self._computeG(q, node.index, radius_id)
                        self._computeF(q, node.index, radius_id)
            processed_nodes += 1
            if processed_nodes % 100 == 0:
                parnas_logger.debug(f'\tProcessed {processed_nodes} nodes')
        parnas_logger.debug(f'Finished DP {datetime.now().strftime("%H:%M:%S")}')

        root_ind = self.tree.seed_node.index
        min_index = int(np.argmin(self.G[self.n_c][root_ind]))
        if self.G[self.n_c][root_ind][min_index] < self.G[0][root_ind][self.nleaves - 1]:
            obj_value = self.G[self.n_c][root_ind][min_index]
            radius_id = self.leaf_lists[root_ind, min_index]
            median_names = self.backtrack(root_ind, self.n_c, radius_id)
            # median_names = [node.taxon.label for node in median_nodes]
            parnas_logger.debug(f'Finished Backtracking {datetime.now().strftime("%H:%M:%S")}')
        else:
            obj_value = self.G[0][root_ind][self.nleaves - 1]
            median_names = []
        parnas_logger.debug(f'Optimal objective function: {obj_value}')
        return obj_value, median_names

    def _initialize_lookups(self):
        """
        Constructs sorted node lists for each node.
        Initializes distance_lookup and r_lookup maps + other helper structures.
        TODO: re-implement to assert O(n^2) runtime. Currently O(n^2 logn)!
        """
        for node1 in self.tree.postorder_node_iter():
            node_dist_pairs = []

            # Fill out the children array.
            if not node1.is_leaf():
                left, right = node1.child_nodes()
                self.children[node1.index, 0] = left.index
                self.children[node1.index, 1] = right.index

            # Compute distances from node1 to all other nodes:
            for node2, dist in dfs_tree_traversal(node1):
                if node2.is_leaf():
                    self.distance_lookup[node1.index, node2.index] = dist

            # for node2 in self.tree.find_clades(lambda x: node1.is_parent_of(x), order='postorder'):
            subtree_size = 0
            for node2 in node1.postorder_iter():
                self.is_ancestor[node1.index, node2.index] = True
                if node2.is_leaf():
                    node_dist_pairs.append((node2.index, self.distance_lookup[node1.index, node2.index]))
                    subtree_size += 1
            self.subtree_leaves[node1.index] = subtree_size
            # for node2 in self.tree.find_clades(lambda x: not node1.is_parent_of(x), order='postorder'):
            for node2 in filtered_postorder_iterator(self.tree, lambda v: v is not node1):
                if node2.is_leaf():
                    node_dist_pairs.append((node2.index, self.distance_lookup[node1.index, node2.index]))
            node_dist_pairs.sort(key=lambda r: r[1])  # sort pairs by distance.

            self.leaf_lists[node1.index, :] = np.array([node2_id for node2_id, dist in node_dist_pairs], dtype=np.int32)

            # Mapping a node to its index in the sorted list for node1:
            for i, (node2_id, dist) in enumerate(node_dist_pairs):
                self.index_lookup[node1.index, node2_id] = i
            # self.distance_lookup[node1.index] = dict([(node, j) for j, (node, dist) in enumerate(node_dist_pairs)])

    def _initialize_G_and_F(self, node: Node):
        self.G[0, node.index] = np.full(self.nleaves, self.distance_functions[node].get_dist(np.inf))
        self.G[1, node.index] = np.full(self.nleaves, 0)
        for radius_node in filtered_preorder_iterator(self.tree, subtree_filter=lambda v: v is not node):
            if not radius_node.is_leaf():
                continue  # skip internals.
            radius_node_list_ind = self.index_lookup[node.index, radius_node.index]
            self.F[0, node.index, radius_node_list_ind] = self.distance_functions[node].get_dist(
                self.distance_lookup[node.index, radius_node.index])
            if self.F[0, node.index, radius_node_list_ind]\
                    > self.G[1, node.index, radius_node_list_ind]:
                self.F[1, node.index, radius_node_list_ind] = self.G[1, node.index, radius_node_list_ind]
            else:
                self.F[1, node.index, radius_node_list_ind] = self.F[0, node.index, radius_node_list_ind]

    def _computeG(self, q: int, node_id: int, radius_id: int):
        if q == 0:
            left_id, right_id = self.children[node_id]
            self.G[q, node_id, self.index_lookup[node_id, radius_id]] =\
                self.G[q, left_id, self.index_lookup[left_id, radius_id]] +\
                self.G[q, right_id, self.index_lookup[right_id, radius_id]]
            return
        if node_id == radius_id:
            return  # skip.
        elif not self.is_ancestor[node_id, radius_id]:
            self.G[q, node_id, self.index_lookup[node_id, radius_id]] = self.G[
                q, node_id, self.index_lookup[node_id, radius_id] - 1]
        else:
            left_id, right_id = self.children[node_id]
            if self.is_ancestor[left_id, radius_id]:
                self._computeG_subtree(q, node_id, radius_id, left_id, right_id)
            else:
                self._computeG_subtree(q, node_id, radius_id, right_id, left_id)

    def _computeG_subtree(self, q: int, node_id: int, radius_id: int, c1: int, c2: int):
        c1_size = self.subtree_leaves[c1]
        c2_size = self.subtree_leaves[c2]

        min_q = max(1, q - c2_size)
        max_q = min(c1_size + 1, q + 1)
        if min_q >= max_q:
            return
        mindist = np.min(
            self.G[min_q:max_q, c1, self.index_lookup[c1, radius_id]] +
            np.flip(self.F[(q-max_q)+1:(q-min_q)+1, c2, self.index_lookup[c2, radius_id]])
        )
        # node_distance = self.distance_functions[node](self.distance_lookup[node.index, radius_node.index]) + mindist
        node_distance = mindist  # NOTE: We assume no contribution of internal nodes!

        if self.G[q, node_id, self.index_lookup[node_id, radius_id] - 1] > node_distance:
            self.G[q, node_id, self.index_lookup[node_id, radius_id]] = node_distance
        else:
            self.G[q, node_id, self.index_lookup[node_id, radius_id]] = self.G[
                q, node_id, self.index_lookup[node_id, radius_id] - 1]

    def _computeF(self, q: int, node_id: int, radius_id: int):
        if self.is_ancestor[node_id, radius_id]:
            return  # Skip.
        left_id, right_id = self.children[node_id]
        left_size = self.subtree_leaves[left_id]
        right_size = self.subtree_leaves[right_id]

        min_q = max(0, q - right_size)
        max_q = min(left_size + 1, q + 1)
        if min_q >= max_q:
            return
        mindist = np.min(
            self.F[min_q:max_q, left_id, self.index_lookup[left_id, radius_id]] +
            np.flip(self.F[(q-max_q)+1:(q-min_q)+1, right_id, self.index_lookup[right_id, radius_id]])
        )
        # self.F[q, self.index_lookup[node], self.distance_lookup[node][radius_node]] = min(
        #     self.G[q, self.index_lookup[node], self.distance_lookup[node][radius_node]],
        #     self.distance_functions[node](self.r_lookup[node][radius_node]) + mindist)
        # node_distance = self.distance_functions[node](self.distance_lookup[node.index, radius_node.index]) + mindist

        node_distance = mindist  # NOTE: we assume no contribution of internal nodes!

        if self.G[q, node_id, self.index_lookup[node_id, radius_id]] > node_distance:
            self.F[q, node_id, self.index_lookup[node_id, radius_id]] = node_distance
        else:
            self.F[q, node_id, self.index_lookup[node_id, radius_id]] = self.G[
                q, node_id, self.index_lookup[node_id, radius_id]]

    def backtrack(self, root_index: int, n_c: int, radius_index: int):
        median_ids = []
        medians = []
        self._backtrack_links(True, root_index, n_c, radius_index, median_ids)
        median_ids = set(median_ids)
        for leaf in self.tree.leaf_nodes():
            if leaf.index in median_ids:
                medians.append(leaf.taxon.label)
        return medians

    def _get_DP_by_id(self, node_id, q, radius_id, in_G):
        radius_list_index = self.index_lookup[node_id, radius_id]
        return self._get_DP_by_list_index(node_id, q, radius_list_index, in_G)

    def _get_DP_by_list_index(self, node_id, q, radius_list_index, in_G):
        source = self.G if in_G else self.F
        if radius_list_index < len(source[q, node_id]):
            return source[q, node_id, radius_list_index]
        else:
            return source[q, node_id][-1]

    def _backtrack_links(self, in_G: bool, node_ind: int, q: int, radius_index: int, medians: List[int]):
        backtrack_queue = deque()
        backtrack_queue.append((in_G, node_ind, q, radius_index))
        while backtrack_queue:
            in_G, node_ind, q, radius_index = backtrack_queue.popleft()
            if not in_G and self.index_lookup[node_ind, radius_index] >= len(self.F[q, node_ind]):
                radius_index = self.leaf_lists[node_ind, len(self.F[q, node_ind]) - 1]
            if q <= 0:
                continue
            if self.children[node_ind, 0] < 0:
                # We are at a leaf.
                if in_G:
                    medians.append(node_ind)
                elif q == 1:
                    rad_list_index = self.index_lookup[node_ind, radius_index]
                    if self.F[1][node_ind][rad_list_index] == self.G[1][node_ind][rad_list_index]:
                        medians.append(node_ind)
            else:
                # Internal node.
                if in_G:
                    cur_val = self.G[q][node_ind][self.index_lookup[node_ind, radius_index]]
                    if self.index_lookup[node_ind, radius_index] > 0 and\
                            cur_val == self.G[q][node_ind][self.index_lookup[node_ind, radius_index] - 1]:
                        next_rad_id = self.leaf_lists[node_ind, self.index_lookup[node_ind, radius_index] - 1]
                        backtrack_queue.append((True, node_ind, q, next_rad_id))
                    else:
                        left, right = self.children[node_ind]
                        c1, c2 = (left, right) if (self.is_ancestor[left, radius_index]) else (right, left)
                        c1_size = self.subtree_leaves[c1]
                        c2_size = self.subtree_leaves[c2]
                        for q1 in range(max(1, q - c2_size), min(c1_size + 1, q + 1)):
                            q2 = q - q1
                            ch_val = self.G[q1][c1][self.index_lookup[c1, radius_index]] +\
                                     self._get_DP_by_id(c2, q2, radius_index, False)
                                     # self.F[c2][q2][self.index_lookup[c2, radius_index]]
                            ch_val += 0  # Assuming that the contribution of internal nodes is 0!
                            if ch_val == cur_val:
                                # Follow links into the children.
                                backtrack_queue.append((True, c1, q1, radius_index))
                                backtrack_queue.append((False, c2, q2, radius_index))
                                break
                else:
                    # In F:
                    cur_val = self.F[q][node_ind][self.index_lookup[node_ind, radius_index]]
                    if cur_val == self.G[q][node_ind][self.index_lookup[node_ind, radius_index]]:
                        backtrack_queue.append((True, node_ind, q, radius_index))
                    else:
                        left, right = self.children[node_ind]
                        left_size = self.subtree_leaves[left]
                        right_size = self.subtree_leaves[right]
                        for q1 in range(max(0, q - right_size), min(left_size + 1, q + 1)):
                            q2 = q - q1
                            ch_val = self._get_DP_by_id(left, q1, radius_index, False) +\
                                     self._get_DP_by_id(right, q2, radius_index, False)
                            ch_val += 0  # Assuming that the contribution of internal nodes is 0!
                            if ch_val == cur_val:
                                # Follow links into the children.
                                backtrack_queue.append((False, left, q1, radius_index))
                                backtrack_queue.append((False, right, q2, radius_index))
                                break



# -*- coding: utf-8 -*-
from collections import deque
from typing import Dict, List, Optional, Set

import numpy as np
from dendropy import Tree, Node
from datetime import datetime

from numba import int32, float64, bool_, typed, types
from numba.experimental import jitclass

from .medoid_utils import DistFunction
from .tree_indexer import TreeIndexer
from parnas.logging import parnas_logger
from .fast_pmedian_finder import FastPMedianFinder, PMedianDP


class TreeCoverage(FastPMedianFinder):
    """
    Finding a set of leaves that cover the entire tree (under a fixed radius).

    This implementation uses numba's JIT to achieve better runtimes.
    """

    def __init__(self, tree: Tree):
        super().__init__(tree)

    def find_coverage(self, radius: float, cost_map: Dict[str, float], covered_leaves: Set[str]) -> Optional[List[str]]:
        self.cost_map = cost_map
        self.distance_functions = None
        self.prior_coverage = np.full(self.nleaves, False, dtype=np.bool_)
        for leaf in self.tree.leaf_nodes():
            if leaf.taxon.label in covered_leaves:
                self.prior_coverage[leaf.index] = True
        parnas_logger.debug(f"Prior coverage: {self.prior_coverage}")
        self._initialize_lookups()

        postorder = np.array([node.index for node in self.tree.postorder_node_iter()], dtype=np.int32)
        coverage_jit = CoverageDP(self.children, postorder, self.is_ancestor, self.leaf_lists, self.subtree_leaves,
                                  self.distance_lookup, self.index_lookup, self.nleaves, self.prior_coverage,
                                  self.costs, self.parents, self.edge_lens, self.tree.seed_node.index, radius)
        self.G, self.F = coverage_jit.run_dp()
        # parnas_logger.debug(str(self.G.tolist()))
        # parnas_logger.debug(str(self.F.tolist()))
        parnas_logger.debug(f'Finished DP {datetime.now().strftime("%H:%M:%S")}')

        root_id = self.tree.seed_node.index
        if self.G[root_id][-1] == np.inf:
            parnas_logger.debug('Impossible to cover the full tree given the constraints.')
            return None
        else:
            parnas_logger.debug(f"Optimal # of reps: {self.G[root_id][-1]}")
            # Do backtracking.
            coverage = self.coverage_backtrack(root_id)
            return coverage

    def coverage_backtrack(self, root_id: int):
        coverage_ids = []
        coverage = []
        self._backtrack_links_coverage(True, root_id, self.leaf_lists[root_id, -1], coverage_ids)
        median_ids = set(coverage_ids)
        for leaf in self.tree.leaf_nodes():
            if leaf.index in median_ids:
                coverage.append(leaf.taxon.label)
        return coverage

    def _backtrack_links_coverage(self, in_G: bool, node_id: int, radius_id: int, coverage: List[int]):
        backtrack_queue = deque()
        backtrack_queue.append((in_G, node_id, radius_id))
        while backtrack_queue:
            in_G, node_id, radius_id = backtrack_queue.popleft()
            if self.children[node_id, 0] < 0:
                # We are at a leaf.
                if (in_G and self.G[node_id, self.index_lookup[node_id, radius_id]] > 0) or\
                        (not in_G and self.F[node_id, self.index_lookup[node_id, radius_id]] > 0):
                    coverage.append(node_id)
                # if in_G or radius_id == node_id:
            else:
                # Internal node.
                if in_G:
                    cur_val = self.G[node_id, self.index_lookup[node_id, radius_id]]
                    if self.index_lookup[node_id, radius_id] > 0 and\
                            cur_val == self.G[node_id, self.index_lookup[node_id, radius_id] - 1]:
                        next_rad_id = self.leaf_lists[node_id, self.index_lookup[node_id, radius_id] - 1]
                        backtrack_queue.append((True, node_id, next_rad_id))
                    else:
                        left, right = self.children[node_id]
                        c1, c2 = (left, right) if (self.is_ancestor[left, radius_id]) else (right, left)
                        backtrack_queue.append((True, c1, radius_id))
                        backtrack_queue.append((False, c2, radius_id))
                else:
                    # In F:
                    cur_val = self.F[node_id, self.index_lookup[node_id, radius_id]]
                    if cur_val == self.G[node_id][self.index_lookup[node_id, radius_id]]:
                        backtrack_queue.append((True, node_id, radius_id))
                    else:
                        left, right = self.children[node_id]
                        backtrack_queue.append((False, left, radius_id))
                        backtrack_queue.append((False, right, radius_id))


@jitclass([('children_arr', int32[:, ::1]), ('postorder_arr', int32[::1]), ('is_ancestor_arr', bool_[:, ::1]),
           ('leaf_lists', int32[:, ::1]), ('subtree_leaves', int32[::1]), ('distance_lookup', float64[:, ::1]),
           ('index_lookup', int32[:, ::1]), ('nleaves', int32), ('nnodes', int32), ('prior_coverage', bool_[::1]),
           ('costs', float64[::1]), ('parents', int32[::1]), ('edge_lengths', float64[::1]), ('root_id', int32),
           ('radius', float64), ('G', float64[:, ::1]), ('F', float64[:, ::1])
           ])
class CoverageDP:
    """
    Numba-optimized dynamic programming solution for the tree coverage problem.
    """

    def __init__(self, children_arr, postorder_arr, is_ancestor_arr, leaf_lists, subtree_leaves, distance_lookup,
                 index_lookup, nleaves: int, prior_coverage, costs, parents, edge_lengths, root_id: int, radius: float):
        self.children_arr = children_arr
        self.postorder_arr = postorder_arr
        self.is_ancestor_arr = is_ancestor_arr
        self.leaf_lists = leaf_lists
        self.subtree_leaves = subtree_leaves
        self.distance_lookup = distance_lookup
        self.index_lookup = index_lookup
        self.prior_coverage = prior_coverage
        self.costs = costs
        self.nleaves = nleaves
        self.nnodes = len(postorder_arr)
        self.parents = parents
        self.edge_lengths = edge_lengths
        self.root_id = root_id
        self.radius = radius

        self._initialize_lookups()
        self.G = np.empty((self.nnodes, self.nleaves), dtype=np.float64)
        self.F = np.empty((self.nnodes, self.nleaves), dtype=np.float64)

    def run_dp(self):
        for node_id in self.postorder_arr:
            # Note that a cost of infinity indicates that a node cannot be selected.
            if self._is_leaf(node_id):
                self.G[node_id].fill(1 if self.costs[node_id] < np.inf else np.inf)
                if self.prior_coverage[node_id]:
                    # The leaf is already covered by prior centers.
                    self.F[node_id].fill(0)
                else:
                    for ind, rad_id in enumerate(self.leaf_lists[node_id]):
                        self_coverage = 1 if self.costs[node_id] < np.inf else np.inf
                        if self.costs[rad_id] < np.inf:
                            self.F[node_id, ind] = 0 if (self.distance_lookup[node_id, rad_id] <= self.radius) else self_coverage
                        else:
                            self.F[node_id, ind] = np.inf
            else:
                for ind, rad_id in enumerate(self.leaf_lists[node_id]):
                    if self.is_ancestor_arr[node_id, rad_id]:
                        prior_G = np.inf if ind == 0 else self.G[node_id, ind - 1]
                        if self.is_ancestor_arr[self.children_arr[node_id, 0], rad_id]:
                            c1, c2 = self.children_arr[node_id, 0], self.children_arr[node_id, 1]
                        else:
                            c1, c2 = self.children_arr[node_id, 1], self.children_arr[node_id, 0]
                        self.G[node_id, ind] = min(
                            prior_G,
                            self._get_DP_by_id(c1, rad_id, True) + self._get_DP_by_id(c2, rad_id, False)
                        )
                        self.F[node_id, ind] = self.G[node_id, ind]
                    else:
                        left, right = self.children_arr[node_id]
                        self.G[node_id, ind] = np.inf if ind == 0 else self.G[node_id, ind - 1]
                        if self.costs[rad_id] < np.inf:
                            self.F[node_id, ind] = min(
                                self.G[node_id, ind],
                                self._get_DP_by_id(left, rad_id, False) + self._get_DP_by_id(right, rad_id, False)
                            )
                        else:
                            self.F[node_id, ind] = np.inf
        return self.G, self.F

    def _is_leaf(self, node_id) -> bool:
        return self.children_arr[node_id, 0] < 0

    def _get_DP_by_id(self, node_id, radius_id, in_G):
        radius_list_index = self.index_lookup[node_id, radius_id]
        return self._get_DP_by_list_index(node_id, radius_list_index, in_G)

    def _get_DP_by_list_index(self, node_id, radius_list_index, in_G):
        source = self.G if in_G else self.F
        if radius_list_index < len(source[node_id]):
            return source[node_id][radius_list_index]
        else:
            return source[node_id][-1]

    # Code below is a copy-paste from PMedianDP. jitclass does not support subtyping.
    # TODO: refactor by taking out the methods below from jitclass structure.
    # ----------------------

    def _initialize_lookups(self):
        """
        Constructs sorted node lists for each node.
        Initializes leaf_lists, index_lookup map + other helper structures.
        TODO: re-implement to assert O(n^2) runtime. Currently O(n^2 logn).
        """
        for node1_id in range(self.nnodes):
            node_dist_pairs = typed.List()
            # Compute distances from node1 to all other nodes:
            for node2_id, dist in zip(*self._dfs_tree_traversal(node1_id)):
                if self._is_leaf(node2_id):
                    self.distance_lookup[node1_id, node2_id] = dist

            # for node2 in self.tree.find_clades(lambda x: node1.is_parent_of(x), order='postorder'):
            subtree_size = 0
            for node2_id in self._postorder_iter(node1_id):
                self.is_ancestor_arr[node1_id, node2_id] = True
                if self._is_leaf(node2_id):
                    node_dist_pairs.append((node2_id, self.distance_lookup[node1_id, node2_id]))
                    subtree_size += 1
            self.subtree_leaves[node1_id] = subtree_size
            # for node2 in self.tree.find_clades(lambda x: not node1.is_parent_of(x), order='postorder'):
            for node2_id in self._postorder_iter(self.root_id, node1_id):
                if self._is_leaf(node2_id):
                    node_dist_pairs.append((node2_id, self.distance_lookup[node1_id, node2_id]))
            node_dist_pairs = sorted(node_dist_pairs, key=lambda r: r[1])
            # node_dist_pairs.sort(key=lambda r: r[1])  # sort pairs by distance.

            self.leaf_lists[node1_id, :] = np.array([node2_id for node2_id, dist in node_dist_pairs], dtype=np.int32)

            # Mapping a node to its index in the sorted list for node1:
            for i, (node2_id, dist) in enumerate(node_dist_pairs):
                self.index_lookup[node1_id, node2_id] = i

    def _dfs_tree_traversal(self, start_node_id: int):
        # DFS stacks (node with prev_node) + distance to node:
        node_stack = np.full((self.nnodes, 2), -1, np.int32)
        dist_stack = np.zeros(self.nnodes, np.float64)  # mirrors node_stack.
        top = 0

        nodes = np.empty(self.nnodes, np.int32)  # list of nodes in dfs order.
        dists = np.empty(self.nnodes, np.float64)  # list of distances to nodes in dfs order.
        i = 0
        node_stack[top, 0], node_stack[top, 1] = start_node_id, -1
        dist_stack[top] = 0
        top = 1
        while top > 0:
            node_id, prev_node_id = node_stack[top - 1]
            cur_dist = dist_stack[top - 1]
            top -= 1
            nodes[i], dists[i] = node_id, cur_dist
            i += 1

            neighbors = typed.List()
            if self.children_arr[node_id, 0] >= 0:
                neighbors.append(self.children_arr[node_id, 0])
                neighbors.append(self.children_arr[node_id, 1])
            if self.parents[node_id] >= 0:
                neighbors.append(self.parents[node_id])
            for neighbor_id in neighbors:
                if neighbor_id != prev_node_id:
                    is_parent = self.parents[node_id] == neighbor_id
                    edge_len = self.edge_lengths[node_id] if is_parent else self.edge_lengths[neighbor_id]
                    node_stack[top, 0], node_stack[top, 1] = neighbor_id, node_id
                    dist_stack[top] = cur_dist + edge_len
                    top += 1
        return nodes, dists

    def _postorder_iter(self, start_node_id: int, filter_node_id=None):
        node_stack = np.empty((self.nnodes, 2), dtype=np.int32)
        node_stack[0, 0], node_stack[0, 1] = start_node_id, 0
        top = 1
        while top > 0:
            node_id, state = node_stack[top - 1]
            top -= 1
            if state:
                yield node_id
            elif node_id != filter_node_id:
                node_stack[top, 0], node_stack[top, 1] = node_id, 1
                top += 1
                if not self._is_leaf(node_id):
                    for child_id in [self.children_arr[node_id, 1], self.children_arr[node_id, 0]]:
                        node_stack[top, 0], node_stack[top, 1] = child_id, 0
                        top += 1



# -*- coding: utf-8 -*-
from typing import List, Dict, Tuple
from dendropy import Tree, Node, Taxon
from Bio.Phylo import BaseTree
import math


def annotate_with_closest_centers(tree: Tree, centers: List[str], prior_centers=None, radius=None):
    """
    For each node in the tree finds its closest center and annotates the nodes with the 'center' keyword.
    :param tree: input tree to be annotated
    :param centers: tip labels of the chosen representatives
    :param prior_centers: an optional list of prior centers - closest nodes will be annotated with 'center=-1'.
    :param radius: optional: if provided, only the vertices within the radius are covered by a center.
    """
    all_centers = centers
    if prior_centers:
        all_centers = centers + prior_centers
    closest_centers = find_closest_centers(tree, all_centers)

    for node in tree.nodes():
        closest_center, dist = closest_centers[node]
        if radius and dist > radius:
            continue
        closest_center = all_centers.index(closest_center)
        if closest_center >= len(centers):
            closest_center = -1
        node.annotations.add_new('center', closest_center)


def find_closest_centers(tree: Tree, centers: List[str]) -> Dict[Node, Tuple[str, float]]:
    """
    For each node in the tree finds its closest center.
    A simple O(kn) algorithm where k is the number of centers and n is the number of leaves in the tree.
    :param tree: input tree to be annotated
    :param centers: tip labels of the chosen representatives
    :return: a dict that maps each node to the closest center/leaf and the respective distance on the tree.
    """
    def tree_traversal(node: Node, prev_node: Node, cur_dist: float, distances: dict):
        distances[node] = cur_dist
        neighbors = []
        if node.child_nodes():
            neighbors += node.child_nodes()
        if node.parent_node:
            neighbors.append(node.parent_node)
        for neighbor in neighbors:
            if neighbor is not prev_node:
                is_parent = node.parent_node is neighbor
                edge_len = node.edge_length if is_parent else neighbor.edge_length
                tree_traversal(neighbor, node, cur_dist + edge_len, distances)

    distances_by_center = []
    for center_label in centers:
        start_leaf = tree.find_node_with_taxon_label(center_label)
        distances = {}
        tree_traversal(start_leaf, None, 0, distances)
        distances_by_center.append(distances)

    center_map = {}
    for node in tree.nodes():
        dist_to_centers = [distances[node] for distances in distances_by_center]
        min_dist = min(dist_to_centers)
        closest_center = dist_to_centers.index(min_dist)
        center_map[node] = (centers[closest_center], min_dist)
    return center_map


class DistFunction(object):
    def __init__(self, is_zero, min_dst=None, max_dist=None, weight=None, is_binary=False):
        if weight is not None and weight < 1e-8:
            is_zero = True
        self.is_zero = is_zero
        self.is_binary = is_binary
        if not is_zero:
            assert None not in (min_dst, max_dist, weight)
            assert min_dst <= max_dist
        self.min_dist = min_dst
        self.max_dist = max_dist
        self.weight = weight

    def get_dist(self, dist: float):
        if self.is_zero:
            return 0

        if dist <= self.min_dist:
            return 0
        if self.min_dist > 0 and self.is_binary:
            return self.weight  # binary means that if a leaf is not covered (dist > min_dist), then contribution is the weight.

        mapped_dist = min(dist, self.max_dist) - self.min_dist  # min_dist (radius) is subtracted from the distance.
        # mapped_dist = dist if (dist <= self.max_dist) else self.max_dist
        mapped_dist *= self.weight
        return mapped_dist


def build_distance_functions(tree: Tree, radius=None, is_binary=False, prior_centers=None, fully_excluded=None,
                             taxa_weights: Dict[str, float]=None) -> Dict[Node, DistFunction]:
    if prior_centers:
        # dendropy_tree = Tree.get(data=tree.format(fmt='newick'), schema='newick',
        #                          preserve_underscores=True)  # convert to dendropy.
        closest_centers = find_closest_centers(tree, prior_centers)
        # For each leaf label stores the distance to the closest prior center:
        closest_prior_dist = dict([(leaf.taxon.label, closest_centers[leaf][1]) for leaf in tree.leaf_nodes()])

    distance_functions = {}
    for node in tree.preorder_node_iter():
        if node.is_leaf():
            excluded = (node.taxon.label in fully_excluded) if fully_excluded else False  # If excluded, dist function is 0.
            weight = taxa_weights.get(node.taxon.label, 1) if taxa_weights else 1  # default weight is 1.
            max_dist = closest_prior_dist[node.taxon.label] if prior_centers else math.inf
            min_dist = radius if radius else 0
            if excluded or max_dist <= min_dist:
                function = DistFunction(is_zero=True)
            else:
                function = DistFunction(False, min_dist, max_dist, weight, is_binary)
        else:
            function = DistFunction(is_zero=True)  # non-terminal nodes do not contribute to the objective function.

        distance_functions[node] = function

    return distance_functions


def get_costs(tree: Tree, excluded=None, fully_excluded=None) -> Dict[str, float]:
    cost_map = {}
    for taxon in tree.taxon_namespace:
        if (excluded and taxon.label in excluded) or (fully_excluded and taxon.label in fully_excluded):
            cost_map[taxon.label] = math.inf
        else:
            cost_map[taxon.label] = 0
    return cost_map


def binarize_tree(tree: Tree, edge_length=0):
    """
    Adds/removes nodes from the tree to make it fully binary (added edges will have length 'edge_length')
    :param tree: Dendropy tree to be made bifurcating.
    """

    # First suppress unifurcations.
    tree.suppress_unifurcations()

    # Now binarize multifurcations.
    for node in tree.postorder_node_iter():
        assert isinstance(node, Node)
        if node.child_nodes() and len(node.child_nodes()) > 2:
            num_children = len(node.child_nodes())
            children = node.child_nodes()
            interim_node = node
            # Creates a caterpillar structure with children on the left of the trunk:
            for child_ind in range(len(children) - 2):
                new_node = Node(edge_length=edge_length)
                interim_node.set_child_nodes([children[child_ind], new_node])
                interim_node = new_node
            interim_node.set_child_nodes(children[num_children - 2:])



# -*- coding: utf-8 -*-
from typing import Callable, Iterable, Optional, List, Tuple
from dendropy import Tree, Node


def dfs_tree_traversal(node: Node) -> List[Tuple[Node, float]]:
    node_dist_pairs = []
    node_stack = []  # node with prev_node and distance.
    node_stack.append((node, None, 0))
    while node_stack:
        node, prev_node, cur_dist = node_stack.pop()
        node_dist_pairs.append((node, cur_dist))
        neighbors = []
        if node.child_nodes():
            neighbors += node.child_nodes()
        if node.parent_node:
            neighbors.append(node.parent_node)
        for neighbor in neighbors:
            if neighbor is not prev_node:
                is_parent = node.parent_node is neighbor
                edge_len = node.edge_length if is_parent else neighbor.edge_length
                node_stack.append((neighbor, node, cur_dist + edge_len))
    return node_dist_pairs


def filtered_preorder_iterator(tree: Tree, subtree_filter: Optional[Callable[[Node], bool]]) -> Iterable[Node]:
    """
    Preorder iterator that will avoid subtrees rooted at nodes, s.t. subtree_filter(node) is False.
    """
    stack = [tree.seed_node]
    while stack:
        node = stack.pop()
        if subtree_filter is None or subtree_filter(node):
            yield node
            stack.extend(n for n in reversed(node._child_nodes))


def filtered_postorder_iterator(tree: Tree, subtree_filter: Optional[Callable[[Node], bool]]) -> Iterable[Node]:
    stack = [(tree.seed_node, False)]
    while stack:
        node, state = stack.pop()
        if state:
            if subtree_filter is None or subtree_filter(node):
                yield node
        elif subtree_filter is None or subtree_filter(node):
            stack.append((node, True))
            stack.extend([(n, False) for n in reversed(node._child_nodes)])


def cost_function():
    """ Returns the setup cost c_j  for a node
    :returns: Number 

    """
    return 0


def distance_function(distance):
    """ Returns the value of distance function f_j
        distance -- the distance between two nodes

    :returns: Number

    """
    return distance


# def resolve_ties(nodelist, treem):  # sort according to condition in paper
#     for i in range(len(nodelist)):
#         if nodelist[i - 1][1] == nodelist[i][1]:  # checks for equidistant vertices vk and vm where vk appears before vm
#             if treem.common_ancestor(nodelist[i - 1][0], nodelist[0][0]) == nodelist[0][0]:  # check if vk is in subtree induced by vj
#                 if treem.common_ancestor(nodelist[i][0], nodelist[0][0]) == nodelist[0][0]:  # check if vm is in subtree induced by vj
#                     if treem.common_ancestor(nodelist[i - 1][0], nodelist[i][0]) == nodelist[i][0]:  ##IF vm is the parent of vk then switch postion of vk and vm
#                         temp = nodelist[i]
#                         nodelist[i] = nodelist[i - 1]
#                         nodelist[i - 1] = temp
#
#             if treem.common_ancestor(nodelist[i][0], nodelist[0][0]) == nodelist[0][0]:  # check if vm is in subtree induced by vj
#                 if treem.common_ancestor(nodelist[i - 1][0], nodelist[0][0]) != nodelist[0][0]:  # if vk is in not in induced by vj
#                     # then switch postion of vk and vm
#                     temp = nodelist[i]
#                     nodelist[i] = nodelist[i - 1]
#                     nodelist[i - 1] = temp
#
#     return nodelist


def calculate_distance(u, v, tree):
    """ Calculates the distance between two nodes
        u -- the node from which distance is calculated
        v -- the node from which distance is to be calculated
        :returns: Number 
    """
    return tree.distance(u, v)



# import numpy as np
# from dendropy import Tree
# from Bio import Phylo
# from io import StringIO
#
# def get_tree(filepath=None,input_string=None):
#     ## TODO add parameters for different input schemas
#     """ Returns a tree based on the input
#
#     Keyword argument:
#     filepath -- the path to the tree file (default None)
#     string -- a string containing the tree
#
#
#     :returns: Dendropy.Tree
#
#     """
#     if filepath:
#         return Tree.get(path=filepath,schema="newick")
#     elif input_string:
#         return Tree.get(data=input_string,schema="newick")
#     else:
#         return Tree()
#
#
# def get_Tree_Phylo(filepath=None,input_string=None):
#     if filepath:
#         return Phylo.parse(filepath,"newick")
#     elif input_string:
#         return Phylo.read(StringIO(input_string),"newick")
#     else:
#         return Tree()
#
# if __name__ == "__main__":
#     t1 = get_tree(input_string="((A,B),(C,(D,E)));") ## TODO add a test input file
#     print(t1.as_ascii_plot())



# -*- coding: utf-8 -*-
from collections import deque
from typing import Dict, List, Optional

import numpy as np
from dendropy import Tree, Node
from datetime import datetime

from numba import int32, float64, bool_, typed, types
from numba.experimental import jitclass

from .medoid_utils import DistFunction
from .tree_indexer import TreeIndexer
from parnas.logging import parnas_logger


class FastPMedianFinder(object):
    """
    Class that finds the median nodes (among leaves!) given a phylogenetic tree.
    An adapted implementation of Tamir's algorithm (Tamir 1996) for generalized p-medians.

    This implementation actively uses numba's JIT to achieve better runtimes.
    Additionally, it is more memory- and time-efficient by cutting redundancies in the DP matrix.
    """

    def __init__(self, tree: Tree):
        tree_indexer = TreeIndexer(tree.taxon_namespace)
        tree_indexer.index_tree(tree)  # Now all nodes in the tree have an 'index' field. First n indices = leaves.

        self.nodes = list(tree.nodes())
        self.nnodes = len(self.nodes)
        self.nleaves = len(tree.leaf_nodes())
        self.tree = tree

        # All arrays bellow will be indexed by node ids.
        # Distance from a node to a leaf:
        self.distance_lookup = np.zeros((self.nnodes, self.nleaves), dtype=np.float64)
        # Index of a leaf in the node's sorted list:
        self.index_lookup = np.zeros((self.nnodes, self.nleaves), dtype=np.int32)
        # Is the first node an ancestor of the second node:
        self.is_ancestor = np.full((self.nnodes, self.nnodes), False, dtype=np.bool_)
        # Number of leaves in the subtree rooted at each node:
        self.subtree_leaves = np.zeros(self.nnodes, dtype=np.int32)
        # Ids of two children nodes for each node ([-1, -1] if a leaf):
        self.children = np.full((self.nnodes, 2), -1, dtype=np.int32)
        # Ids of parents nodes and lengths of parent edges:
        self.parents = np.full(self.nnodes, -1, dtype=np.int32)
        self.edge_lens = np.zeros(self.nnodes, dtype=np.float64)
        # Leaf lists for each node sorted by distance (+ additional properties - see Tamir 1996):
        self.leaf_lists = np.zeros((self.nnodes, self.nleaves), dtype=np.int32)
        # Costs array:
        self.costs = np.zeros(self.nleaves, dtype=np.float64)

        self.n_c = len(self.tree.leaf_nodes())
        self.G = np.array([], dtype=np.float64)
        self.F = np.array([], dtype=np.float64)

    def find_medoids(self, p: int, distance_functions: Dict[Node, DistFunction], cost_map: Dict[str, float]):
        parnas_logger.debug(f'Started find_medoids {datetime.now().strftime("%H:%M:%S")}')
        self.distance_functions = distance_functions
        self.cost_map = cost_map
        # Parameters of a distance function for each leaf stored in an array:
        self.dist_func_array = np.empty((self.nleaves, 5), dtype=np.float64)
        # Number of medoids to choose:
        self.n_c = min(p, len(self.tree.leaf_nodes()))

        # Initializes all the key structures (bounded by O(n^2) space) needed for DP:
        self._initialize_lookups()
        parnas_logger.debug(f'Finished initialization {datetime.now().strftime("%H:%M:%S")}')

        postorder = np.array([node.index for node in self.tree.postorder_node_iter()], dtype=np.int32)
        pmedian_jit = PMedianDP(self.children, postorder, self.is_ancestor, self.leaf_lists, self.subtree_leaves,
                                self.distance_lookup, self.index_lookup, self.n_c, self.nleaves, self.dist_func_array,
                                self.costs, self.parents, self.edge_lens, self.tree.seed_node.index)
        self.G, self.F = pmedian_jit.run_dp()
        parnas_logger.debug(f'Finished DP {datetime.now().strftime("%H:%M:%S")}')

        root_ind = self.tree.seed_node.index
        min_index = int(np.argmin(self.G[root_ind][self.n_c]))
        if self.G[root_ind][self.n_c][min_index] < self.G[root_ind][0][self.nleaves - 1]:
            obj_value = self.G[root_ind][self.n_c][min_index]
            radius_id = self.leaf_lists[root_ind, min_index]
            median_names = self.backtrack(root_ind, self.n_c, radius_id)
            # median_names = [node.taxon.label for node in median_nodes]
            parnas_logger.debug(f'Finished Backtracking {datetime.now().strftime("%H:%M:%S")}')
        else:
            # No new medoids are really needed in this case.
            obj_value = self.G[root_ind][0][self.nleaves - 1]
            median_names = []

        parnas_logger.debug(f'Optimal objective function: {obj_value}')
        return obj_value, median_names

    def get_score(self, k: int) -> Optional[float]:
        root_ind = self.tree.seed_node.index
        if not self.G or len(self.G) <= 0 or k >= len(self.G[root_ind]):
            return None
        else:
            return self.G[root_ind][k][-1]

    def _is_leaf(self, node_id):
        return self.children[node_id, 0] < 0

    def _initialize_lookups(self):
        for node1 in self.tree.postorder_node_iter():
            node_dist_pairs = []
            if node1.parent_node:
                self.edge_lens[node1.index] = node1.edge_length
                self.parents[node1.index] = node1.parent_node.index

            # Fill out dist_func and costs array.
            if node1.is_leaf():
                self.costs[node1.index] = self.cost_map[node1.taxon.label]
                if self.distance_functions:
                    func = self.distance_functions[node1]
                    if func.is_zero:
                        self.dist_func_array[node1.index, 0] = 0
                    else:
                        self.dist_func_array[node1.index, 0] = 1
                        self.dist_func_array[node1.index, 1:] = [func.min_dist, func.max_dist, func.weight,
                                                                 1 if func.is_binary else 0]

            # Fill out the children array.
            if not node1.is_leaf():
                left, right = node1.child_nodes()
                self.children[node1.index, 0] = left.index
                self.children[node1.index, 1] = right.index

            # # Compute distances from node1 to all other nodes:
            # for node2, dist in dfs_tree_traversal(node1):
            #     if node2.is_leaf():
            #         self.distance_lookup[node1.index, node2.index] = dist
            #
            # # for node2 in self.tree.find_clades(lambda x: node1.is_parent_of(x), order='postorder'):
            # subtree_size = 0
            # for node2 in node1.postorder_iter():
            #     self.is_ancestor[node1.index, node2.index] = True
            #     if node2.is_leaf():
            #         node_dist_pairs.append((node2.index, self.distance_lookup[node1.index, node2.index]))
            #         subtree_size += 1
            # self.subtree_leaves[node1.index] = subtree_size
            # # for node2 in self.tree.find_clades(lambda x: not node1.is_parent_of(x), order='postorder'):
            # for node2 in filtered_postorder_iterator(self.tree, lambda v: v is not node1):
            #     if node2.is_leaf():
            #         node_dist_pairs.append((node2.index, self.distance_lookup[node1.index, node2.index]))
            # node_dist_pairs.sort(key=lambda r: r[1])  # sort pairs by distance.
            #
            # self.leaf_lists[node1.index, :] = np.array([node2_id for node2_id, dist in node_dist_pairs], dtype=np.int32)
            #
            # # Mapping a node to its index in the sorted list for node1:
            # for i, (node2_id, dist) in enumerate(node_dist_pairs):
            #     self.index_lookup[node1.index, node2_id] = i
            # # self.distance_lookup[node1.index] = dict([(node, j) for j, (node, dist) in enumerate(node_dist_pairs)])

    def backtrack(self, root_index: int, n_c: int, radius_index: int):
        median_ids = []
        medians = []
        self._backtrack_links(True, root_index, n_c, radius_index, median_ids)
        median_ids = set(median_ids)
        for leaf in self.tree.leaf_nodes():
            if leaf.index in median_ids:
                medians.append(leaf.taxon.label)
        return medians

    def _get_DP_by_id(self, node_id, q, radius_id, in_G):
        radius_list_index = self.index_lookup[node_id, radius_id]
        return self._get_DP_by_list_index(node_id, q, radius_list_index, in_G)

    def _get_DP_by_list_index(self, node_id, q, radius_list_index, in_G):
        source = self.G if in_G else self.F
        if radius_list_index < len(source[node_id][q]):
            return source[node_id][q][radius_list_index]
        else:
            return source[node_id][q][-1]

    def _backtrack_links(self, in_G: bool, node_ind: int, q: int, radius_index: int, medians: List[int]):
        backtrack_queue = deque()
        backtrack_queue.append((in_G, node_ind, q, radius_index))
        while backtrack_queue:
            in_G, node_ind, q, radius_index = backtrack_queue.popleft()
            if not in_G and self.index_lookup[node_ind, radius_index] >= len(self.F[node_ind][q]):
                radius_index = self.leaf_lists[node_ind, len(self.F[node_ind][q]) - 1]
            if q <= 0:
                continue
            if self.children[node_ind, 0] < 0:
                # We are at a leaf.
                if in_G:
                    medians.append(node_ind)
                elif q == 1:
                    rad_list_index = self.index_lookup[node_ind, radius_index]
                    if self.F[node_ind][1][rad_list_index] == self.G[node_ind][1][rad_list_index]:
                        medians.append(node_ind)
            else:
                # Internal node.
                if in_G:
                    cur_val = self._get_DP_by_id(node_ind, q, radius_index, True)
                    #cur_val = self.G[node_ind][q][self.index_lookup[node_ind, radius_index]]
                    if self.index_lookup[node_ind, radius_index] > 0 and\
                            cur_val == self._get_DP_by_list_index(node_ind, q, self.index_lookup[node_ind, radius_index] - 1, True):
                            # cur_val == self.G[node_ind][q][self.index_lookup[node_ind, radius_index] - 1]:
                        next_rad_id = self.leaf_lists[node_ind, self.index_lookup[node_ind, radius_index] - 1]
                        backtrack_queue.append((True, node_ind, q, next_rad_id))
                    else:
                        left, right = self.children[node_ind]
                        c1, c2 = (left, right) if (self.is_ancestor[left, radius_index]) else (right, left)
                        c1_size = self.subtree_leaves[c1]
                        c2_size = self.subtree_leaves[c2]
                        for q1 in range(max(1, q - c2_size), min(c1_size + 1, q + 1)):
                            q2 = q - q1
                            ch_val = self._get_DP_by_id(c1, q1, radius_index, True) +\
                                     self._get_DP_by_id(c2, q2, radius_index, False)
                                     # self.G[c1][q1][self.index_lookup[c1, radius_index]] +\
                                     # self.F[c2][q2][self.index_lookup[c2, radius_index]]
                            ch_val += 0  # Assuming that the contribution of internal nodes is 0!
                            if ch_val == cur_val:
                                # Follow links into the children.
                                backtrack_queue.append((True, c1, q1, radius_index))
                                backtrack_queue.append((False, c2, q2, radius_index))
                                break
                else:
                    # In F:
                    cur_val = self._get_DP_by_id(node_ind, q, radius_index, False)
                    # cur_val = self.F[node_ind][q][self.index_lookup[node_ind, radius_index]]
                    # if cur_val == self.G[node_ind][q][self.index_lookup[node_ind, radius_index]]:
                    if cur_val == self._get_DP_by_id(node_ind, q, radius_index, True):
                        backtrack_queue.append((True, node_ind, q, radius_index))
                    else:
                        left, right = self.children[node_ind]
                        left_size = self.subtree_leaves[left]
                        right_size = self.subtree_leaves[right]
                        for q1 in range(max(0, q - right_size), min(left_size + 1, q + 1)):
                            q2 = q - q1
                            ch_val = self._get_DP_by_id(left, q1, radius_index, False) +\
                                     self._get_DP_by_id(right, q2, radius_index, False)
                            ch_val += 0  # Assuming that the contribution of internal nodes is 0!
                            if ch_val == cur_val:
                                # Follow links into the children.
                                backtrack_queue.append((False, left, q1, radius_index))
                                backtrack_queue.append((False, right, q2, radius_index))
                                break


@jitclass([('children_arr', int32[:, ::1]), ('postorder_arr', int32[::1]), ('is_ancestor_arr', bool_[:, ::1]),
           ('leaf_lists', int32[:, ::1]), ('subtree_leaves', int32[::1]), ('distance_lookup', float64[:, ::1]),
           ('index_lookup', int32[:, ::1]), ('n_c', int32), ('nleaves', int32), ('nnodes', int32),
           ('dist_func_arr', float64[:, ::1]), ('costs', float64[::1]), ('parents', int32[::1]),
           ('edge_lengths', float64[::1]), ('root_id', int32),
           ('G', types.ListType(types.ListType(float64[::1]))), ('F', types.ListType(types.ListType(float64[::1])))
           ])
class PMedianDP:
    """
    Numba-optimized dynamic programming solution for the p-medians (k-medoids) problem.
    Note that for numba typed.List data structure accessing elements by getitem_unchecked() achieves a 10-15 times
    faster performance than typical [] access.
    """

    def __init__(self, children_arr, postorder_arr, is_ancestor_arr, leaf_lists, subtree_leaves,
                 distance_lookup, index_lookup, n_c: int, nleaves: int, dist_func_arr, costs, parents, edge_lengths,
                 root_id: int):
        self.children_arr = children_arr
        self.postorder_arr = postorder_arr
        self.is_ancestor_arr = is_ancestor_arr
        self.leaf_lists = leaf_lists
        self.subtree_leaves = subtree_leaves
        self.distance_lookup = distance_lookup
        self.index_lookup = index_lookup
        self.dist_func_arr = dist_func_arr
        self.costs = costs
        self.n_c = n_c
        self.nleaves = nleaves
        self.nnodes = len(postorder_arr)
        self.parents = parents
        self.edge_lengths = edge_lengths
        self.root_id = root_id

        self._initialize_lookups()

        # This is a necessary hack to properly initialize typed lists by filling them with proper content.
        # We allocate arrays for q=0 for all nodes in the tree (q > 0 arrays will be allocated dynamically later):
        g_temp = typed.List()
        f_temp = typed.List()
        for i in range(self.nnodes):
            gq0 = typed.List()
            gq0.append(np.full(self.nleaves, np.inf, dtype=np.float64))
            g_temp.append(gq0)
            fq0 = typed.List()
            fq0.append(np.full(self.nleaves, np.inf, dtype=np.float64))
            f_temp.append(fq0)
        self.G = g_temp
        self.F = f_temp

    def run_dp(self):
        for node_id in self.postorder_arr:
            G_node = self.G.getitem_unchecked(node_id)
            F_node = self.F.getitem_unchecked(node_id)
            if self._is_leaf(node_id):
                # Allocate and append arrays for q = 1 (we do not need q > 1):
                G_node.append(np.full(self.nleaves, self.costs[node_id], dtype=np.float64))
                F_node.append(np.full(self.nleaves, np.inf, dtype=np.float64))
                self._initialize_G_and_F(node_id)
            else:
                max_q = min(self.subtree_leaves[node_id], self.n_c)
                # self.G[node_id].append(np.full(1, np.inf, dtype=np.float64))
                # self.F[node_id]np.full(self.nleaves, np.inf, dtype=np.float64)
                max_radius = self.nleaves - 1
                for q in range(1, max_q + 1):
                    while not self.is_ancestor_arr[node_id, self.leaf_lists[node_id, max_radius]]:
                        max_radius -= 1
                    G_node.append(np.full(max_radius + 1, np.inf, dtype=np.float64))
                    F_node.append(np.full(max_radius + 1, np.inf, dtype=np.float64))
                    if self.distance_lookup[node_id, self.leaf_lists[node_id, max_radius]] >\
                            self.distance_lookup[node_id, self.leaf_lists[node_id, max_radius - 1]]:
                        max_radius -= 1

                for q in range(max_q + 1):
                    for radius_list_index in range(len(G_node.getitem_unchecked(q))):
                        self._computeG(q, node_id, self.leaf_lists[node_id, radius_list_index])
                        self._computeF(q, node_id, self.leaf_lists[node_id, radius_list_index])
        return self.G, self.F

    def _is_leaf(self, node_id) -> bool:
        return self.children_arr[node_id, 0] < 0

    def _dist_func(self, node_id: int, dist: float):
        if self.dist_func_arr[node_id, 0] < 0.5:
            return 0
        else:
            min_dist, max_dist, weight, is_binary = self.dist_func_arr[node_id, 1:]
            if dist <= min_dist:
                return 0
            if min_dist > 0 and is_binary > 0.5:
                return weight
            # mapped_dist = dist if (dist <= max_dist) else max_dist  # This is an older interpretation -- obsolete.
            mapped_dist = min(dist, max_dist) - min_dist
            mapped_dist = mapped_dist * weight
            return mapped_dist

    def _get_DP_by_id(self, node_id, q, radius_id, in_G):
        radius_list_index = self.index_lookup[node_id, radius_id]
        return self._get_DP_by_list_index(node_id, q, radius_list_index, in_G)

    def _get_DP_by_list_index(self, node_id, q, radius_list_index, in_G):
        source = self.G if in_G else self.F
        dp_row = source.getitem_unchecked(node_id).getitem_unchecked(q)
        if radius_list_index < len(dp_row):
            return dp_row[radius_list_index]
        else:
            return dp_row[-1]

    def _initialize_lookups(self):
        """
        Constructs sorted node lists for each node.
        Initializes leaf_lists, index_lookup map + other helper structures.
        TODO: re-implement to assert O(n^2) runtime. Currently O(n^2 logn).
        """
        for node1_id in range(self.nnodes):
            node_dist_pairs = typed.List()
            # Compute distances from node1 to all other nodes:
            for node2_id, dist in zip(*self._dfs_tree_traversal(node1_id)):
                if self._is_leaf(node2_id):
                    self.distance_lookup[node1_id, node2_id] = dist

            # for node2 in self.tree.find_clades(lambda x: node1.is_parent_of(x), order='postorder'):
            subtree_size = 0
            for node2_id in self._postorder_iter(node1_id):
                self.is_ancestor_arr[node1_id, node2_id] = True
                if self._is_leaf(node2_id):
                    node_dist_pairs.append((node2_id, self.distance_lookup[node1_id, node2_id]))
                    subtree_size += 1
            self.subtree_leaves[node1_id] = subtree_size
            # for node2 in self.tree.find_clades(lambda x: not node1.is_parent_of(x), order='postorder'):
            for node2_id in self._postorder_iter(self.root_id, node1_id):
                if self._is_leaf(node2_id):
                    node_dist_pairs.append((node2_id, self.distance_lookup[node1_id, node2_id]))
            node_dist_pairs = sorted(node_dist_pairs, key=lambda r: r[1])
            # node_dist_pairs.sort(key=lambda r: r[1])  # sort pairs by distance.

            self.leaf_lists[node1_id, :] = np.array([node2_id for node2_id, dist in node_dist_pairs], dtype=np.int32)

            # Mapping a node to its index in the sorted list for node1:
            for i, (node2_id, dist) in enumerate(node_dist_pairs):
                self.index_lookup[node1_id, node2_id] = i

    def _initialize_G_and_F(self, node_id: int):
        self.G.getitem_unchecked(node_id)[0] = np.full(self.nleaves,  self._dist_func(node_id, np.inf))
        # self.G[node_id] = np.full(self.nleaves, 0)
        for radius_id in self.postorder_arr:
            if not self._is_leaf(radius_id) or self.is_ancestor_arr[node_id, radius_id]:
                continue  # skip internals and descendants.
            radius_node_list_ind = self.index_lookup[node_id, radius_id]
            # self.F[0, node_id, radius_node_list_ind] = self.distance_functions[node](
            #     self.distance_lookup[node.index, radius_node.index])
            self.F[node_id][0][radius_node_list_ind] = self._dist_func(node_id, self.distance_lookup[node_id, radius_id])
            self.F[node_id][1][radius_node_list_ind] = min(self.F[node_id][0][radius_node_list_ind],
                                                           self.G[node_id][1][radius_node_list_ind])

    def _computeG(self, q: int, node_id: int, radius_id: int):
        node_q_row = self.G.getitem_unchecked(node_id).getitem_unchecked(q)
        if q == 0:
            left_id, right_id = self.children_arr[node_id]
            node_q_row[self.index_lookup[node_id, radius_id]] =\
                self.G.getitem_unchecked(left_id).getitem_unchecked(0)[self.index_lookup[left_id, radius_id]] +\
                self.G.getitem_unchecked(right_id).getitem_unchecked(0)[self.index_lookup[right_id, radius_id]]
            return
        if node_id == radius_id:
            return  # skip.
        elif not self.is_ancestor_arr[node_id, radius_id]:
            if self.index_lookup[node_id, radius_id] > 0:
                node_q_row[self.index_lookup[node_id, radius_id]] =\
                    node_q_row[self.index_lookup[node_id, radius_id] - 1]
            else:
                node_q_row[self.index_lookup[node_id, radius_id]] = np.inf
        else:
            left_id, right_id = self.children_arr[node_id]
            if self.is_ancestor_arr[left_id, radius_id]:
                self._computeG_subtree(q, node_id, radius_id, left_id, right_id)
            else:
                self._computeG_subtree(q, node_id, radius_id, right_id, left_id)

    def _computeG_subtree(self, q: int, node_id: int, radius_id: int, c1: int, c2: int):
        c1_size = self.subtree_leaves[c1]
        c2_size = self.subtree_leaves[c2]

        min_q = max(1, q - c2_size)
        max_q = min(c1_size + 1, q + 1)
        if min_q >= max_q:
            return

        mindist = np.inf
        for q1 in range(min_q, max_q):
            q2 = q - q1
            dist = self._get_DP_by_id(c1, q1, radius_id, True) + self._get_DP_by_id(c2, q2, radius_id, False)
            # print(str(dist))
            # dist = self.G[c1.index][q1][self.index_lookup[c1.index, radius_node.index]] +\
            #        self.F[q2, c2.index, self.index_lookup[c2.index, radius_node.index]]
            if dist < mindist:
                mindist = dist
        # print('------')
        # node_distance = self.distance_functions[node](self.distance_lookup[node.index, radius_node.index]) + mindist
        node_distance = mindist  # NOTE: We assume no contribution of internal nodes!

        if self.index_lookup[node_id, radius_id] - 1 >= 0:
            self.G.getitem_unchecked(node_id).getitem_unchecked(q)[self.index_lookup[node_id, radius_id]] = min(
                node_distance,
                self._get_DP_by_list_index(node_id, q, self.index_lookup[node_id, radius_id] - 1, True)
            )
        else:
            self.G.getitem_unchecked(node_id).getitem_unchecked(q)[
                self.index_lookup[node_id, radius_id]] = node_distance

    def _computeF(self, q: int, node_id: int, radius_id: int):
        if self.is_ancestor_arr[node_id, radius_id]:
            # return  # Skip.
            self.F.getitem_unchecked(node_id).getitem_unchecked(q)[self.index_lookup[node_id, radius_id]] =\
                self.G.getitem_unchecked(node_id).getitem_unchecked(q)[self.index_lookup[node_id, radius_id]]
        left_id, right_id = self.children_arr[node_id]
        left_size = self.subtree_leaves[left_id]
        right_size = self.subtree_leaves[right_id]

        min_q = max(0, q - right_size)
        max_q = min(left_size + 1, q + 1)
        if min_q >= max_q:
            return

        mindist = np.inf
        for q1 in range(min_q, max_q):
            q2 = q - q1
            dist = self._get_DP_by_id(left_id, q1, radius_id, False) + self._get_DP_by_id(right_id, q2, radius_id, False)
            # dist = self.F[q1, left.index, self.index_lookup[left.index, radius_node.index]] + self.F[
            #     q2, right.index, self.index_lookup[right.index, radius_node.index]]
            if dist< mindist:
                mindist = dist

        node_distance = mindist  # NOTE: we assume no contribution of internal nodes!

        self.F.getitem_unchecked(node_id).getitem_unchecked(q)[self.index_lookup[node_id, radius_id]] = min(
            node_distance,
            self._get_DP_by_id(node_id, q, radius_id, True)
        )

    def _dfs_tree_traversal(self, start_node_id: int):
        # DFS stacks (node with prev_node) + distance to node:
        node_stack = np.full((self.nnodes, 2), -1, np.int32)
        dist_stack = np.zeros(self.nnodes, np.float64)  # mirrors node_stack.
        top = 0

        nodes = np.empty(self.nnodes, np.int32)  # list of nodes in dfs order.
        dists = np.empty(self.nnodes, np.float64)  # list of distances to nodes in dfs order.
        i = 0
        node_stack[top, 0], node_stack[top, 1] = start_node_id, -1
        dist_stack[top] = 0
        top = 1
        while top > 0:
            node_id, prev_node_id = node_stack[top - 1]
            cur_dist = dist_stack[top - 1]
            top -= 1
            nodes[i], dists[i] = node_id, cur_dist
            i += 1

            neighbors = typed.List()
            if self.children_arr[node_id, 0] >= 0:
                neighbors.append(self.children_arr[node_id, 0])
                neighbors.append(self.children_arr[node_id, 1])
            if self.parents[node_id] >= 0:
                neighbors.append(self.parents[node_id])
            for neighbor_id in neighbors:
                if neighbor_id != prev_node_id:
                    is_parent = self.parents[node_id] == neighbor_id
                    edge_len = self.edge_lengths[node_id] if is_parent else self.edge_lengths[neighbor_id]
                    node_stack[top, 0], node_stack[top, 1] = neighbor_id, node_id
                    dist_stack[top] = cur_dist + edge_len
                    top += 1
        return nodes, dists

    def _postorder_iter(self, start_node_id: int, filter_node_id=None):
        node_stack = np.empty((self.nnodes, 2), dtype=np.int32)
        node_stack[0, 0], node_stack[0, 1] = start_node_id, 0
        top = 1
        while top > 0:
            node_id, state = node_stack[top - 1]
            top -= 1
            if state:
                yield node_id
            elif node_id != filter_node_id:
                node_stack[top, 0], node_stack[top, 1] = node_id, 1
                top += 1
                if not self._is_leaf(node_id):
                    for child_id in [self.children_arr[node_id, 1], self.children_arr[node_id, 0]]:
                        node_stack[top, 0], node_stack[top, 1] = child_id, 0
                        top += 1

# @jit('f8(i4,i4,i4,i4,i4,i4,i4,f8[:,:,:],f8[:,:,:])')
# def jit_G_DP(q, left_id, right_id, left_rad, right_rad, min_q, max_q, G, F):
#     mindist = np.inf
#     for q1 in range(min_q, max_q):
#         q2 = q - q1
#         dist = G[q1, left_id, left_rad] + F[q2, right_id, right_rad]
#         if dist < mindist:
#             mindist = dist
#     return mindist
#
#
# @jit('f8(i4,i4,i4,i4,i4,i4,i4,f8[:,:,:])')
# def jit_F_DP(q, left_id, right_id, left_rad, right_rad, min_q, max_q, F):
#     mindist = np.inf
#     for q1 in range(min_q, max_q):
#         q2 = q - q1
#         dist = F[q1, left_id, left_rad] + F[q2, right_id, right_rad]
#         if dist < mindist:
#             mindist = dist
#     return mindist



# -*- coding: utf-8 -*-
from dendropy import TaxonNamespace, Tree


class InvalidArgumentError(Exception):
    def __init__(self, name: str, value: str, message=''):
        super(InvalidArgumentError, self).__init__('Invalid argument %s: %s; %s' % (name, value, message))
        self.name = name
        self.value = value


class TreeIndexer:
    """
    Adds 'index' field to all nodes and taxa on the passed trees.
    It ensures that the taxa are indexed consistently across trees.
    """

    def __init__(self, taxon_namespace: TaxonNamespace):
        self.label_mapping = {}
        index = 0
        for taxon in taxon_namespace:
            self.label_mapping[taxon.label] = index
            index += 1

    def index_tree(self, tree: Tree):
        for leaf_node in tree.leaf_nodes():
            if leaf_node.taxon.label in self.label_mapping:
                leaf_node.taxon.index = self.label_mapping[leaf_node.taxon.label]
            else:
                print(leaf_node.taxon.label)
                print(self.label_mapping)
                raise InvalidArgumentError('tree', '', 'Input tree should be over the initially specified taxon set')

        node_id = 0
        for leaf in tree.leaf_node_iter():
            leaf.index = node_id
            node_id += 1
        for node in tree.preorder_internal_node_iter():
            node.index = node_id
            # node.annotations.add_new('id', node_id)
            node_id += 1



# -*- coding: utf-8 -*-

import colorsys
from datetime import datetime
from typing import List
import sys
# import math

# from Bio import SeqRecord
from dendropy import Tree

from parnas import parnas_logger
from parnas.options import parser, parse_and_validate
# from parnas.sequences import SequenceSimilarityMatrix
from parnas.medoids import find_n_medoids, annotate_with_closest_centers, build_distance_functions, binarize_tree,\
    get_costs, find_n_medoids_with_diversity, find_coverage


# os.environ["NUMBA_DUMP_ANNOTATION"] = "1"
sys.setrecursionlimit(100000)


def color_by_clusters(tree: Tree, centers: List[str], prior_centers=None, fully_excluded=None, radius=None):
    # Annotate the tree nodes with the closest centers.
    annotate_with_closest_centers(tree, centers, prior_centers=prior_centers, radius=radius)

    # Choose n different colors. If prior_centers list is not empty, add an additional reserved color.
    n = len(centers) if not prior_centers else len(centers) + 1
    hsv_colors = [(i / n, 0.7, 0.7) for i in range(n)]
    rgb_colors = [colorsys.hsv_to_rgb(*hsv) for hsv in hsv_colors]
    rgb_colors = [[round(255 * color[i]) for i in range(len(color))] for color in rgb_colors]
    hex_colors = ['#%02x%02x%02x' % (color[0], color[1], color[2]) for color in rgb_colors]
    if prior_centers:
        prior_color = hex_colors[0]  # Reserve the first color for prior centers.
        hex_colors = hex_colors[1:]

    # Color the edges.
    for edge in tree.preorder_edge_iter():
        if edge.head_node:
            edge.head_node.annotations.drop(name='!color')  # remove any previous color scheme.
        # If the two ends of the edge are covered by the same center -- color it.
        if edge.head_node and edge.tail_node and edge.head_node.annotations.get_value('center') is not None and\
                edge.head_node.annotations.get_value('center') == edge.tail_node.annotations.get_value('center'):
            if edge.head_node.annotations.get_value('center') < 0:  # negative if it is covered by a prior center.
                edge.head_node.annotations.add_new('!color', prior_color)  # use the reserved color.
            else:
                center_ind = edge.head_node.annotations.get_value('center')
                color = hex_colors[center_ind]
                edge.head_node.annotations.add_new('!color', color)

    # Drop previous color annotations and color 'regular' taxa black (to avoid color mixing in FigTree).
    black = '#000000'
    grey = '#a9a9a9'
    special_taxa = centers
    if prior_centers:
        special_taxa = special_taxa + prior_centers
    special_taxa = set(special_taxa)
    for taxon in tree.taxon_namespace:
        taxon.annotations.drop(name='!color')  # Remove any previous colors if any.
        if not taxon.label in special_taxa:
            if fully_excluded and taxon.label in fully_excluded:
                taxon.annotations.add_new('!color', grey)
            else:
                taxon.annotations.add_new('!color', black)

    # Color the centers.
    for center_ind, center in enumerate(centers):
        taxon = tree.taxon_namespace.get_taxon(center)
        color = hex_colors[center_ind]
        taxon.annotations.add_new('!color', color)

    # Color prior centers if applicable.
    if prior_centers:
        for prior_center in prior_centers:
            taxon = tree.taxon_namespace.get_taxon(prior_center)
            taxon.annotations.add_new('!color', prior_color)


# def assess_clade_similarity(annotated_tree: Tree, centers: List[str], records: List[SeqRecord.SeqRecord],
#                             sim_matrix: SequenceSimilarityMatrix):
#     for i, center in enumerate(centers):
#         print('Clade of %s:' % center)
#         max_within_dist = 0
#         min_outside_dist = math.inf
#         for leaf1 in annotated_tree.leaf_nodes():
#             if leaf1.taxon.label != center:
#                 continue
#             for leaf2 in annotated_tree.leaf_nodes():
#                 leaf1_ind = [i for i, rec in enumerate(records) if rec.id == leaf1.taxon.label][0]
#                 leaf2_ind = [i for i, rec in enumerate(records) if rec.id == leaf2.taxon.label][0]
#                 dist = 1 - sim_matrix.matrix[leaf1_ind][leaf2_ind]
#                 if leaf1.annotations.get_value('center') == i and leaf2.annotations.get_value('center') == i:
#                     if dist > max_within_dist:
#                         max_within_dist = dist
#                 if leaf1.annotations.get_value('center') == i and leaf2.taxon.label in centers and leaf1 != leaf2:
#                     if dist < min_outside_dist:
#                         min_outside_dist = dist
#         print('\tmax within clade: %.3f%%' % (max_within_dist * 100))
#         print('\tclosest other representative: %.3f%%' % (min_outside_dist * 100))


def run_parnas_cli():
    args, query_tree, n, radius, is_binary, prior_centers, excluded_taxa, obj_excluded, fully_excluded, taxa_weights = parse_and_validate()

    # Binarize the query tree:
    binarize_tree(query_tree, edge_length=0)

    dist_functions = build_distance_functions(query_tree, prior_centers=prior_centers, is_binary=is_binary,
                                              fully_excluded=fully_excluded + obj_excluded, radius=radius,
                                              taxa_weights=taxa_weights)
    cost_map = get_costs(query_tree, excluded_taxa, fully_excluded)
    parnas_logger.info("Inferring best representatives...")
    if not args.cover:
        representatives, value, diversity_scores = find_n_medoids_with_diversity(query_tree, n, dist_functions, cost_map,
                                                                                 max_dist=radius)
    else:
        # coverage = None
        coverage = find_coverage(query_tree, radius, cost_map, prior_centers, fully_excluded, obj_excluded)
        if coverage is None:
            parnas_logger.warning("The tree cannot be fully covered given the exclusion constraints.")
            parnas_logger.warning("Falling back onto a slower method that would cover the tree as much as possible.")
            opt_n = -1
            prev_value = -1
            for n in range(1, len(query_tree.leaf_nodes()) + 1):
                representatives, value = find_n_medoids(query_tree, n, dist_functions, cost_map, max_dist=radius)
                if value == prev_value:
                    # The best possible value was achieved on the previous n (full coverage is impossible).
                    opt_n = n - 1
                    break
                elif value == 0:
                    # Achieved full coverage.
                    opt_n = n
                    break
                prev_value = value
        else:
            representatives = coverage

    if len(representatives) == 0:
        parnas_logger.info('The diversity on the tree is already fully covered by the prior centers - no new '
                           'representatives needed.')
    else:
        parnas_logger.info('Chosen representatives:')
        for rep in representatives:
            print('%s' % rep)
    if not args.cover and len(representatives) > 1 and diversity_scores:
        parnas_logger.info('Chosen representatives cover %.2f%% of ' % diversity_scores[-1] +
                           f'{"(new) " if prior_centers else "overall "}diversity.')

    if args.out_path:
        color_by_clusters(query_tree, representatives, prior_centers=prior_centers, fully_excluded=fully_excluded,
                          radius=radius)
        parnas_logger.debug(f'Finished coloring {datetime.now().strftime("%H:%M:%S")}')
        try:
            query_tree.write(path=args.out_path, schema='nexus')
            parnas_logger.info('Colored tree was saved to "%s".' % args.out_path)
        except Exception:
            parser.error('Cant write to the specified path "%s".' % args.out_path)

    if args.csv_path:
        if args.cover:
            parnas_logger.warning('"--diversity" cannot be used in combination with "--cover".')
        elif n <= 1:
            parnas_logger.warning('No diversity scores to report as n < 2.')
        elif diversity_scores:
            try:
                with open(args.csv_path, 'w') as diversity_log:
                    diversity_log.write('Representatives, Diversity_covered\n')
                    for i, score in enumerate(diversity_scores):
                        diversity_log.write('%d, %.2f\n' % (i + 2, score))
                parnas_logger.info('Diversity scores were saved to "%s".' % args.csv_path)
            except IOError:
                parser.error('Cant write to the specified path "%s".' % args.csv_path)

    if args.sample_tree_path:
        sample_tree = query_tree.extract_tree_with_taxa_labels(representatives)
        assert isinstance(sample_tree, Tree)
        sample_tree.purge_taxon_namespace()
        for taxon in sample_tree.taxon_namespace:
            taxon.annotations.drop(name='!color')
        sample_tree.write(path=args.sample_tree_path, schema='nexus')

        # if args.nt_alignment:
        #     # Parsing assuming its swIAV HA sequences.
        #     alignment_path = args.nt_alignment
        #     records = list(SeqIO.parse(alignment_path, 'fasta'))
        #     sim_matrix = SequenceSimilarityMatrix(None, None, records, records, aligned=True, ignore_tails=True, nt_alphabet=True)
        #     assess_clade_similarity(query_tree, representatives, records, sim_matrix)



from setuptools import setup

from parnas.version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    install_requires=[
        'numpy<=1.24,>=1.17',
        'numba>=0.57.0',
        'biopython>=1.67',
        'dendropy>=4.5.0',
        'phylo-treetime>=0.9.4'
    ],
    name="parnas",
    version=__version__,
    author="Alexey Markin, Sanket Wagle, Siddhant Grover",
    author_email="alex.markin57@gmail.com",
    license='MIT',
    description="Representative taxon sampling from phylogenetic trees",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/flu-crew/parnas",
    packages=["parnas", "parnas.medoids"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": ["parnas=parnas.cli:run_parnas_cli"]},
    py_modules=["parnas"],
)



# -*- coding: utf-8 -*-
import argparse
import os
import re
import subprocess
from argparse import RawTextHelpFormatter, RawDescriptionHelpFormatter
from math import floor

from Bio import AlignIO
from Bio.Align import MultipleSeqAlignment
from dendropy import Tree

from parnas import parnas_logger

# Program interface:
parser = argparse.ArgumentParser(description='Phylogenetic mAximum RepreseNtAtion Sampling (PARNAS)',
                                 formatter_class=RawDescriptionHelpFormatter)
parser._optionals.title = "Arguments"
parser.add_argument('-t', '--tree', type=str, action='store', dest='tree',
                    help='Path to the input tree in newick or nexus format.', required=True)
parser.add_argument('-n', type=int, action='store', dest='samples',
                    help='Number of representatives to be chosen.\n' +
                         'This argument is required unless the --cover option is specified')
parser.add_argument('--prior', type=str, action='store', dest='prior_regex',
                    help='Indicate the previous representatives (if any) with a regex. '
                         'The regex should match a full taxon name.\n'
                         'PARNAS will then select centers that represent diversity '
                         'not covered by the previous representatives.', required=False)
# parser.add_argument('--threshold', type=float, action='store', dest='percent',
#                     help='sequences similarity threshold: the algorithm will choose best representatives that cover as much\n' +
#                          'diversity as possible within the given similarity threshold. ' +
#                          '--nt or --aa must be specified with this option', required=False)
parser.add_argument('--weights', type=str, action='store', dest='weights_csv',
                    help='A CSV file specifying a weight for some or all taxa. '
                         'The column names must be "taxon" and "weight".\n'
                         'If a taxon is not listed in the file, its weight is assumed to be 1. '
                         'Maximum allowed weight is 1000 and weights below 1e-8 are considered 0.')
parser.add_argument('--radius', type=float, action='store', dest='radius',
                    help='Each representative will "cover" all leaves within the specified radius on the tree. '
                         'PARNAS will then choose representatives so that the amount of uncovered diversity is minimized.',
                    required=False)
parser.add_argument('--cover', action='store_true',
                    help="Choose the best representatives (smallest number) that cover all the tips within the specified radius/threshold.\n" +
                    "If specified, a --radius or --threshold argument must be specified as well.",
                    required=False)
parser.add_argument('--binary', action='store_true',
                    help="To be used with --radius. Instead of covering as much diversity as possible, "
                         "PARNAS will cover as many tips as possible within the radius. "
                         "Each leaf will have a binary contribution to the objective: 0 if covered, else its weight.",
                    required=False)

output_options = parser.add_argument_group('Output options')
output_options.add_argument('--color', type=str, action='store', dest='out_path',
                    help='PARNAS will save a colored tree, where the chosen representatives are highlighted '
                    'and the tree is color-partitioned respective to the representatives.\n'
                    'If prior centers are specified, they (and the subtrees they represent) will be colored red.')
output_options.add_argument('--diversity', type=str, action='store', dest='csv_path',
                    help='Save diversity scores for all k (number of representatives) from 2 to n.\n'
                         'Can be used to choose the right number of representatives for a dataset.')
output_options.add_argument('--subtree', type=str, action='store', dest='sample_tree_path',
                    help='Prune the tree to the sampled taxa and save to the specified file in NEXUS format.')

taxa_handler = parser.add_argument_group('Excluding taxa')
taxa_handler.add_argument('--exclude-rep', type=str, action='store', dest='exclude_regex',
                          help='Prohibits parnas to choose representatives from the taxa matching this regex. '
                               'However, the excluded taxa will still contribute to the objective function.')
taxa_handler.add_argument('--exclude-obj', type=str, action='store', dest='eobj_regex',
                          help='Matching taxa can be selected, but will not contribute to the objective function. '
                               'Can be used if one wants to select taxa from a reference set.')
taxa_handler.add_argument('--exclude-fully', type=str, action='store', dest='full_regex',
                          help='Completely ignore the taxa matching this regex.')
taxa_handler.add_argument('--constrain-fully', type=str, action='store', dest='constrain_regex',
                          help='Completely ignore the taxa NOT matching this regex.')

alignment_parser = parser.add_argument_group('Controlling sequence divergence')
alignment_parser.add_argument('--threshold', type=float, action='store', dest='percent',
                              help='The sequence similarity threshold that works as --radius. ' +
                                   'For example, "95" will imply that each representative ' +
                                   'covers all leaves within 5%% divergence on the tree.\n'
                                   'To account for sequence divergence, parnas will run TreeTime to infer ancestral substitutions '
                                   'along the tree edges and re-weigh the edges based on the number of sustitutions on them.\n'
                                   'A sequence alignment (--nt or --aa) must be specified with this option',
                              required=False)
alignment_parser.add_argument('--nt', type=str, action='store', dest='nt_alignment',
                    help='Path to nucleotide sequence alignment associated with the tree tips.', required=False)
alignment_parser.add_argument('--aa', type=str, action='store', dest='aa_alignment',
                    help='Path to amino acid sequence alignment associated with the tree tips.', required=False)
# parser.add_argument('--prior', metavar='TAXON', type=str, nargs='+',
#                     help='space-separated list of taxa that have been previously chosen as centers.\n' +
#                          'The algorithm will choose new representatives that cover the "new" diversity in the tree')


# Computes the coverage radius (# of substitutions) that satisfies the similarity threshold.
def threshold_to_substitutions(sim_threshold: float, alignment: MultipleSeqAlignment) -> int:
    subs = floor((1 - sim_threshold / 100) * len(alignment[0]))
    parnas_logger.info("%.3f%% similarity threshold implies that a single representative will cover all tips "
                       "in the %d-substitution radius." % (sim_threshold, subs))
    return subs


def reweigh_tree_ancestral(tree_path: str, alignment_path: str, aa=False) -> Tree:
    """
    Re-weighs the tree edges according to the number of substitutions per edge.
    The ancestral substitutions are inferred using TreeTime (Sargulenko et al. 2018).
    :param tree_path: path to the tree to be re-weighed.
    :param alignment_path: path to MSA associated with the tree tips.
    :param aa: whether the sequences consist of amino acid residues (default: nucleotide).
    :return: The re-weighed tree
    """
    # Run ancestral inference with treetime.
    treetime_outdir = 'treetime_ancestral_%s' % tree_path.split(os.sep)[-1]
    if not os.path.exists(treetime_outdir):
        os.mkdir(treetime_outdir)
    treetime_log_path = '%s/treetime.log' % treetime_outdir
    parnas_logger.info('Inferring ancestral substitutions with TreeTime. The log will be written to "%s".' % treetime_log_path)
    command = ['treetime', 'ancestral', '--aln', alignment_path, '--tree', tree_path, '--outdir', treetime_outdir,
               '--gtr', 'infer']
    if aa:
        command += ['--aa']
    treetime_out = '%s/annotated_tree.nexus' % treetime_outdir
    with open(treetime_log_path, 'w') as treetime_log:
        subprocess.call(command, stdout=treetime_log, stderr=subprocess.STDOUT)

    # Read the treetime output and weight the edges according to the number of subs.
    try:
        # Update the treetime output file.
        treetime_for_dendropy = '%s/ancestral_updated.nexus' % treetime_outdir
        with open(treetime_out, 'r') as treetime_nexus:
            with open(treetime_for_dendropy, 'w') as dendropy_nexus:
                for line in treetime_nexus:
                    for mutations in re.findall(r'mutations=".*?"', line):
                        upd_mutations = mutations.replace(',',
                                                          '||')  # DendroPy does not like commas in the annotations.
                        line = line.replace(mutations, upd_mutations)
                    dendropy_nexus.write(line)
        ancestral_tree = Tree.get(path=treetime_for_dendropy, schema='nexus', preserve_underscores=True)
    except Exception:
        parser.error('Failed to infer an ancestral substitutions with TreeTime. '
                     'Please see the TreeTime output log.')

    parnas_logger.info('Re-weighing the tree based on ancestral substitutions.')
    reweighed_tree_path = '%s/ancestral_reweighed.tre' % treetime_outdir
    reweighed_tree = ancestral_tree
    for node in reweighed_tree.nodes():
        edge_length = 0
        mutations_str = node.annotations.get_value('mutations')
        if mutations_str and mutations_str.strip():
            edge_length = mutations_str.count('||') + 1
        node.edge_length = edge_length
    reweighed_tree.write(path=reweighed_tree_path, schema='newick')
    return reweighed_tree


def validate_weights(path: str):
    taxa_weights = {}
    try:
        with open(path, 'r') as weights_csv:
            headers = [header.strip() for header in weights_csv.readline().split(',')]
            if len(headers) != 2 or headers[0] != 'taxon' or headers[1] != 'weight':
                parser.error('Invalid weights file: the first line of the scv file must be "taxon, weight".')
            while True:
                l = weights_csv.readline().strip()
                if not l:
                    break
                taxon, weight_str = [value.strip() for value in l.split(',')]
                weight = float(weight_str)
                if weight < 0 or weight > 1000:
                    parser.error('Invalid weight %.8f for taxon %s. All weights must be non-negative and not exceed 1000.'
                                 % (weight, taxon))
                taxa_weights[taxon] = weight
    except Exception:
        parser.error('Cannot open/read the specified weights file "%s".' % path)
    return taxa_weights


def find_matching_taxa(tree: Tree, regex: str, title: str, none_message: str, print_taxa=True):
    matching_taxa = []
    for taxon in tree.taxon_namespace:
        if re.match(regex, taxon.label):
            matching_taxa.append(taxon.label)

    if print_taxa:
        if matching_taxa:
            parnas_logger.info(title)
            for t in matching_taxa:
                parnas_logger.plain('\t%s' % t)
        else:
            parnas_logger.info(none_message)
        parnas_logger.plain('')
    return matching_taxa


def parse_and_validate():
    args = parser.parse_args()

    # Validate the tree.
    tree = None
    try:
        tree = Tree.get(path=args.tree, schema='newick', preserve_underscores=True)
    except Exception as newick_exception:
        try:
            tree = Tree.get(path=args.tree, schema='nexus', preserve_underscores=True)
        except Exception as nexus_exception:
            parser.error('Cannot read the specified tree file "%s". ' % args.tree +
                         'Make sure the tree is in the newick or nexus format.\n'
                         f'Nexus error: {nexus_exception}\n'
                         f'Newick error: {newick_exception}')

    # Validate n.
    n = -1
    if not args.samples and not args.cover:
        parser.error('Please either specify the number of representatives with "-n" or use the --cover option.')
    if args.samples:
        n = args.samples
        if n < 1 or n >= len(tree.taxon_namespace):
            parser.error('n should be at least 1 and smaller than the number of taxa in the tree.')

    # Handle --prior-regex.
    prior_centers = None
    if args.prior_regex:
        prior_centers = find_matching_taxa(tree, args.prior_regex, 'Prior centers that match the regex:',
                                           'No taxa matched PRIOR_REGEX', True)

    # Validate exclusions.
    excluded_taxa = []
    fully_excluded = []
    obj_excluded = []
    if args.exclude_regex:
        excluded_taxa = find_matching_taxa(tree, args.exclude_regex,
                                           'Not considering the following as representatives (matched EXCLUDE_REGEX):',
                                           'No taxa matched EXCLUDE_REGEX', True)
    if args.full_regex:
        fully_excluded = find_matching_taxa(tree, args.full_regex, 'Ignoring the following taxa (matched FULL_REGEX):',
                                            'No taxa matched FULL_REGEX', True)
    if args.eobj_regex:
        obj_excluded = find_matching_taxa(tree, args.eobj_regex,
                                          'Not contributing to the objective (matched EOBJ_REGEX):',
                                          'No taxa matched EOBJ_REGEX')
    if args.constrain_regex:
        constrained_taxa = find_matching_taxa(tree, args.constrain_regex,
                                              'Constraining to the following taxa (matched CONSTRAIN_REGEX):',
                                              'No taxa matched CONSTRAIN_REGEX', True)
        fully_excluded += list({taxon.label for taxon in tree.taxon_namespace}.symmetric_difference(constrained_taxa))

    exclude_intersection = set(excluded_taxa).intersection(set(fully_excluded))
    if exclude_intersection:
        for taxon in exclude_intersection:
            parnas_logger.warning(f'{taxon} matches both EXCLUDE_REGEX and FULL_REGEX. PARNAS will fully exclude it.')

    # Validate radius.
    radius = None
    if args.radius is not None:
        if args.radius <= 0:
            parser.error('radius should be positive.')
        else:
            radius = args.radius

    # Validate alignment.
    if args.nt_alignment or args.aa_alignment:
        if args.nt_alignment and args.aa_alignment:
            parser.error('Please specify EITHER the nucleotide or amino acid alignment - not both.')
        alignment_path = args.nt_alignment if args.nt_alignment else args.aa_alignment
        is_aa = args.aa_alignment is not None
        try:
            alignment = list(AlignIO.read(alignment_path, 'fasta'))
        except Exception:
            parser.error('Cannot read the specified FASTA alignment in "%s".' % alignment_path)
        alignment_present = True
    else:
        alignment_present = False

    # Validate threshold-related parameters and re-weigh the tree
    if args.percent:
        if args.percent <= 0 or args.percent >= 100:
            parser.error('Invalid "--threshold %.3f" option. The threshold must be between 0 and 100 (exclusive)'
                         % args.percent)
        if not alignment_present:
            parser.error('To use the --threshold parameter, please specify a nucleotide' +
                         'or amino acid alignment associated with the tree tips.')
        else:
            radius = threshold_to_substitutions(args.percent, alignment)
            query_tree = reweigh_tree_ancestral(args.tree, alignment_path, is_aa)
    else:
        query_tree = tree

    is_binary = args.binary
    if is_binary and not radius:
        is_binary = False

    # Validate cover
    if args.cover:
        if not args.percent and not args.radius:
            parser.error('To use --cover parameter, please specify --threshold or --radius option.')

    # Validate weights.
    taxa_weights = None
    if args.weights_csv:
        taxa_weights = validate_weights(args.weights_csv)

    return args, query_tree, n, radius, is_binary, prior_centers, excluded_taxa, obj_excluded, fully_excluded, taxa_weights



# -*- coding: utf-8 -*-
from typing import List, Tuple, Dict, Optional
from dendropy import Tree

from .fast_pmedian_finder import FastPMedianFinder
from .pmedian_finder import PMedianFinder
from .medoid_utils import find_closest_centers
from .tree_coverage import TreeCoverage
from ..logging import parnas_logger


def find_n_medoids(tree: Tree, n: int, distance_functions: Dict, cost_map: Dict[str, float], max_dist=None)\
        -> Tuple[List[str], float]:
    """
    Finds n medoids on a tree by a modification of Tamir's algorithm for p-median.
    If max_dist is specified, the method finds n representatives that cover
    as much diversity (within max_dist) as possible.

    :param tree: phylogenetic tree in the dendropy format.
    :param n: number of representatives to be chosen.
    :param distance_functions: a map that links a node (dendropy.Node) to the distance function of that node.
    :param max_dist: an optional parameter that specifies the maximum coverage distance by a single representative.
    :return: (1) a list of tip labels that have been chosen as representatives;
             (2) the minimal objective function value.
    """
    medoidFinder = FastPMedianFinder(tree)
    objective, medoids = medoidFinder.find_medoids(n, distance_functions, cost_map)

    return medoids, objective


def find_n_medoids_with_diversity(tree: Tree, n: int, distance_functions: Dict, cost_map: Dict[str, float], max_dist=None)\
        -> Tuple[List[str], float, List[float]]:
    """
    Mirrors "find_n_medoids", but adds a diversity score to the output.
    """
    medoidFinder = FastPMedianFinder(tree)
    objective, medoids = medoidFinder.find_medoids(n, distance_functions, cost_map)
    diversity_scores = []
    if n > 1:
        obj1 = medoidFinder.get_score(1)  # Objective function value for n=1 (1 medoid).
        if obj1 > 0:
            for k in range(2, n + 1):
                objk = medoidFinder.get_score(k)
                diversity_score = (obj1 - objk) / obj1 * 100  # Calculate the % of covered diversity.
                diversity_scores.append(diversity_score)
    return medoids, objective, diversity_scores


def find_coverage(tree: Tree, radius: float, cost_map: Dict[str, float], prior_centers: List[str] = None,
                  fully_excluded: List[str] = None, obj_excluded: List[str] = None) -> Optional[List[str]]:
    """
    Find an optimal set of taxa that fully cover all leaves within the specified radius.
    """
    tree_copy = tree.clone()
    labels_to_prune = []
    if fully_excluded:
        labels_to_prune.extend(fully_excluded)
    # Find leaves that are already covered by prior centers:
    covered_leaves = set()
    if prior_centers:
        closest_priors = find_closest_centers(tree, prior_centers)
        for node, (prior_center, dist) in closest_priors.items():
            if dist <= radius and node.taxon:
                # labels_to_prune.append(node.taxon.label)  # Can't prune covered leaves -- they can be used as reps.
                covered_leaves.add(node.taxon.label)
    if obj_excluded:  # Also ignore taxa that are excluded from the objective
        for label in obj_excluded:
            covered_leaves.add(label)
    # Prune all fully excluded leaves:
    if labels_to_prune:
        if len(labels_to_prune) == len(tree_copy.leaf_nodes()):
            return []
        else:
            tree_copy.prune_taxa_with_labels(labels_to_prune)
            # parnas_logger.debug("After pruning:" + tree_copy.as_string("newick"))
    if covered_leaves and len(covered_leaves) == len(tree_copy.leaf_nodes()):
        return []

    tree_coverer = TreeCoverage(tree_copy)
    coverage = tree_coverer.find_coverage(radius, cost_map, covered_leaves)  # Compute the coverage.
    return coverage


# if __name__ == "__main__":
#    tree = get_Tree_Phylo(input_string="((A:2,B:3):4,(C:5,(D:7,E:1):7):11);")
#    #tree = get_Tree_Phylo(input_string="((A:23,B:27):47,(C:35,(D:76,E:18):28):31);")
#
#    print(find_n_medoids(tree, 3, None))



# -*- coding: utf-8 -*-

from .tree_medoids import find_n_medoids, find_n_medoids_with_diversity, find_coverage
from .medoid_utils import annotate_with_closest_centers, build_distance_functions, binarize_tree, get_costs
